from typing import ClassVar

from engin import Block, Invoke, Provide, Supply, provide
from examples.fastapi.routes.cats.adapters.repository import InMemoryCatRepository
from examples.fastapi.routes.cats.api import router
from examples.fastapi.routes.cats.ports import CatRepository


class CatBlock(Block):
    options: ClassVar[list[Provide | Invoke]] = [Supply([router])]

    @provide
    def cat_repository(self) -> CatRepository:
        return InMemoryCatRepository()



from typing import Annotated

from engin import Provide
from engin._dependency import Supply
from tests.deps import make_aliased_int


def test_provide_discriminates_singular():
    def i_provide() -> int:
        return 3

    provider = Provide(i_provide)
    assert not provider.is_multiprovider
    assert provider.return_type_id.type is int
    assert not provider.return_type_id.multi


def test_provide_discriminates_multi():
    def i_provide_multi() -> list[int]:
        return [3]

    provider = Provide(i_provide_multi)
    assert provider.is_multiprovider
    assert provider.return_type_id.type is int
    assert provider.return_type_id.multi


def test_supply_singular():
    class Foo: ...

    supply = Supply(Foo())
    assert not supply.is_multiprovider
    assert supply.return_type_id.type is Foo
    assert not supply.return_type_id.multi


def test_supply_multi():
    class Foo: ...

    supply = Supply([Foo()])
    assert supply.return_type_id.type is Foo
    assert supply.return_type_id.multi
    assert supply.is_multiprovider


def test_provide_with_alias():
    provider = Provide(make_aliased_int)

    assert provider.return_type_id.type


def test_provide_with_annotation():
    def make_str_1() -> Annotated[str, "1"]:
        return "bar"

    provider = Provide(make_str_1)

    assert provider.return_type_id.type
    assert str(provider.return_type_id) == "Annotated[str, 1]"



from typing import Any

from engin._dependency import Provide


class ProviderError(Exception):
    """
    Raised when a Provider errors during Assembly.
    """

    def __init__(
        self,
        provider: Provide[Any],
        error_type: type[Exception],
        error_message: str,
    ) -> None:
        self.provider = provider
        self.error_type = error_type
        self.error_message = error_message
        self.message = (
            f"provider '{provider.name}' errored with error "
            f"({error_type.__name__}): '{error_message}'"
        )

    def __str__(self) -> str:
        return self.message



import logging

import uvicorn

from engin import Supply
from engin.ext.asgi import ASGIEngin
from examples.asgi.app import AppBlock, AppConfig
from examples.asgi.common.db.block import DatabaseBlock
from examples.asgi.features.cats.block import CatBlock

logging.basicConfig(level=logging.DEBUG)

app = ASGIEngin(AppBlock(), DatabaseBlock(), CatBlock(), Supply(AppConfig(debug=True)))


if __name__ == "__main__":
    uvicorn.run(app)



from typing import ClassVar

from pydantic import ValidationError
from starlette.requests import Request
from starlette.responses import Response

from examples.asgi.common.db.ports import DatabaseInterface
from examples.asgi.common.starlette.endpoint import Endpoint
from examples.asgi.features.cats.domain import Cat


class PostCatEndpoint(Endpoint):
    ALLOWED_METHODS: ClassVar[list[str]] = ["POST"]

    def __init__(self, db: DatabaseInterface) -> None:
        self._db = db

    async def exec(self, request: Request) -> Response:
        cat_dto = await request.json()

        try:
            cat = Cat.model_validate(cat_dto)
        except ValidationError:
            return Response(status_code=422)

        self._db.set(cat.name, cat.model_dump())

        return Response(status_code=204)



from collections.abc import Callable
from typing import Annotated, TypeAlias

from engin._type_utils import TypeId, type_id_of


def test_type_id_of_int():
    type_id = type_id_of(int)
    assert type_id == TypeId(type=int, multi=False)
    assert str(type_id) == "int"


def test_type_id_of_list_of_int():
    type_id = type_id_of(list[int])
    assert type_id == TypeId(type=int, multi=True)
    assert str(type_id) == "int[]"


def test_type_id_of_alias():
    MyAlias: TypeAlias = int

    assert (
        type_id_of(MyAlias)
        == TypeId(type=MyAlias, multi=False)
        == TypeId(type=int, multi=False)
    )


def test_type_id_of_type_of_class():
    class Foo: ...

    type_id = type_id_of(type[Foo])
    assert type_id == TypeId(type=type[Foo], multi=False)
    assert str(type_id) == "type[Foo]"


def test_type_id_of_callable():
    func = Callable[[int], str]
    type_id = type_id_of(func)
    assert type_id == TypeId(type=Callable[[int], str], multi=False)
    assert str(type_id) == "Callable[[int], str]"


def test_type_id_of_multi_callable():
    many_func = list[Callable[[int], str]]
    type_id = type_id_of(many_func)
    assert type_id == TypeId(type=Callable[[int], str], multi=True)
    assert str(type_id) == "Callable[[int], str][]"


def test_type_id_of_annotation():
    annotated = Annotated[int, "Activity"]
    type_id = type_id_of(annotated)
    assert type_id == TypeId(type=Annotated[int, "Activity"], multi=False)
    assert str(type_id) == "Annotated[int, Activity]"


def test_type_id_of_complex_annotation():
    annotated = Annotated[Callable[[int], str], "Activity"]
    type_id = type_id_of(annotated)
    assert type_id == TypeId(type=Annotated[Callable[[int], str], "Activity"], multi=False)
    assert str(type_id) == "Annotated[Callable[[int], str], Activity]"



from typing import ClassVar

from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from examples.asgi.common.db.ports import DatabaseInterface
from examples.asgi.common.starlette.endpoint import Endpoint
from examples.asgi.features.cats.domain import Cat


class GetCatEndpoint(Endpoint):
    ALLOWED_METHODS: ClassVar[list[str]] = ["GET"]

    def __init__(self, db: DatabaseInterface) -> None:
        self._db = db

    async def exec(self, request: Request) -> Response:
        name = request.path_params.get("name")
        if name is None:
            cat_dtos = self._db.list()
            return JSONResponse(
                content=[Cat.model_validate(cat_dto).model_dump() for cat_dto in cat_dtos]
            )

        cat_dto = self._db.get(name)

        if cat_dto is None:
            return Response(status_code=404)

        cat = Cat.model_validate(cat_dto)

        return JSONResponse(content=cat.model_dump_json())



import inspect
import typing
from abc import ABC
from collections.abc import Awaitable, Callable
from inspect import Parameter, Signature, isclass, iscoroutinefunction
from typing import (
    Any,
    Generic,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    get_type_hints,
)

from engin._type_utils import TypeId, type_id_of

P = ParamSpec("P")
T = TypeVar("T")
Func: TypeAlias = Callable[P, T]


def _noop(*args: Any, **kwargs: Any) -> None: ...


class Dependency(ABC, Generic[P, T]):
    def __init__(self, func: Func[P, T], block_name: str | None = None) -> None:
        self._func = func
        self._is_async = iscoroutinefunction(func)
        self._signature = inspect.signature(self._func)
        self._block_name = block_name

    @property
    def origin(self) -> str:
        """
        The module that this Dependency originated from.

        Returns:
            A string, e.g. "examples.fastapi.app"
        """
        return self._func.__module__

    @property
    def block_name(self) -> str | None:
        return self._block_name

    @property
    def func_name(self) -> str:
        return self._func.__name__

    @property
    def name(self) -> str:
        if self._block_name:
            return f"{self._block_name}.{self._func.__name__}"
        else:
            return f"{self._func.__module__}.{self._func.__name__}"

    @property
    def parameter_types(self) -> list[TypeId]:
        parameters = list(self._signature.parameters.values())
        if not parameters:
            return []
        if parameters[0].name == "self":
            parameters.pop(0)
        return [type_id_of(param.annotation) for param in parameters]

    @property
    def signature(self) -> Signature:
        return self._signature

    def set_block_name(self, name: str) -> None:
        self._block_name = name

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self._is_async:
            return await cast(Awaitable[T], self._func(*args, **kwargs))
        else:
            return self._func(*args, **kwargs)


class Invoke(Dependency):
    """
    Marks a function as an Invocation.

    Invocations are functions that are called prior to lifecycle startup. Invocations
    should not be long running as the application startup will be blocked until all
    Invocation are completed.

    Invocations can be provided as an Option to the Engin or a Block.

    Examples:
        ```python3
        def print_string(a_string: str) -> None:
            print(f"invoking with value: '{a_string}'")

        invocation = Invoke(print_string)
        ```
    """

    def __init__(self, invocation: Func[P, T], block_name: str | None = None) -> None:
        super().__init__(func=invocation, block_name=block_name)

    def __str__(self) -> str:
        return f"Invoke({self.name})"


class Entrypoint(Invoke):
    """
    Marks a type as an Entrypoint.

    Entrypoints are a short hand for no-op Invocations that can be used to
    """

    def __init__(self, type_: type[Any], *, block_name: str | None = None) -> None:
        self._type = type_
        super().__init__(invocation=_noop, block_name=block_name)

    @property
    def origin(self) -> str:
        return self._type.__module__

    @property
    def parameter_types(self) -> list[TypeId]:
        return [type_id_of(self._type)]

    @property
    def signature(self) -> Signature:
        return Signature(
            parameters=[
                Parameter(name="x", kind=Parameter.POSITIONAL_ONLY, annotation=self._type)
            ]
        )

    def __str__(self) -> str:
        return f"Entrypoint({type_id_of(self._type)})"


class Provide(Dependency[Any, T]):
    def __init__(self, builder: Func[P, T], block_name: str | None = None) -> None:
        super().__init__(func=builder, block_name=block_name)
        self._is_multi = typing.get_origin(self.return_type) is list

        if self._is_multi:
            args = typing.get_args(self.return_type)
            if len(args) != 1:
                raise ValueError(
                    f"A multiprovider must be of the form list[X], not '{self.return_type}'"
                )

    @property
    def return_type(self) -> type[T]:
        if isclass(self._func):
            return_type = self._func  # __init__ returns self
        else:
            try:
                return_type = get_type_hints(self._func, include_extras=True)["return"]
            except KeyError as err:
                raise RuntimeError(
                    f"Dependency '{self.name}' requires a return typehint"
                ) from err

        return return_type

    @property
    def return_type_id(self) -> TypeId:
        return type_id_of(self.return_type)

    @property
    def is_multiprovider(self) -> bool:
        return self._is_multi

    def __hash__(self) -> int:
        return hash(self.return_type_id)

    def __str__(self) -> str:
        return f"Provide({self.return_type_id})"


class Supply(Provide, Generic[T]):
    def __init__(
        self, value: T, *, type_hint: type | None = None, block_name: str | None = None
    ) -> None:
        self._value = value
        self._type_hint = type_hint
        if self._type_hint is not None:
            self._get_val.__annotations__["return"] = type_hint
        super().__init__(builder=self._get_val, block_name=block_name)

    @property
    def origin(self) -> str:
        return self._value.__module__

    @property
    def return_type(self) -> type[T]:
        if self._type_hint is not None:
            return self._type_hint
        if isinstance(self._value, list):
            return list[type(self._value[0])]  # type: ignore[misc,return-value]
        return type(self._value)

    def _get_val(self) -> T:
        return self._value

    def __str__(self) -> str:
        return f"Supply({self.return_type_id})"



import inspect
import typing
from collections.abc import Iterable
from inspect import Parameter
from typing import ClassVar, TypeVar

from fastapi.routing import APIRoute

from engin import Engin, Entrypoint, Invoke, Option
from engin._dependency import Dependency, Supply
from engin._graph import DependencyGrapher, Node
from engin._type_utils import TypeId, type_id_of
from engin.ext.asgi import ASGIEngin

try:
    from fastapi import APIRouter, FastAPI
    from fastapi.params import Depends
    from starlette.requests import HTTPConnection
except ImportError as err:
    raise ImportError(
        "fastapi package must be installed to use the fastapi extension"
    ) from err

__all__ = ["APIRouteDependency", "FastAPIEngin", "Inject"]


def _attach_engin(
    app: FastAPI,
    engin: Engin,
) -> None:
    app.state.engin = engin


class FastAPIEngin(ASGIEngin):
    _LIB_OPTIONS: ClassVar[list[Option]] = [*ASGIEngin._LIB_OPTIONS, Invoke(_attach_engin)]
    _asgi_type = FastAPI

    def graph(self) -> list[Node]:
        grapher = _FastAPIDependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve(
            [
                Entrypoint(self._asgi_type),
                *[i for i in self._invocations if i.func_name != "_attach_engin"],
            ]
        )


T = TypeVar("T")


def Inject(interface: type[T]) -> Depends:
    async def inner(conn: HTTPConnection) -> T:
        engin: Engin = conn.app.state.engin
        return await engin.assembler.get(interface)

    dep = Depends(inner)
    dep.__engin__ = True  # type: ignore[attr-defined]
    return dep


class _FastAPIDependencyGrapher(DependencyGrapher):
    """
    This exists in order to bridge the gap between
    """

    def _resolve_recursive(
        self, roots: Iterable[Dependency], *, seen: set[TypeId]
    ) -> list[Node]:
        nodes: list[Node] = []
        for root in roots:
            for parameter in root.parameter_types:
                provider = self._providers[parameter]

                # multiprovider
                if isinstance(provider, list):
                    for p in provider:
                        nodes.append(Node(node=p, parent=root))

                        if isinstance(p, Supply):
                            route_dependencies = _extract_routes_from_supply(p)
                            nodes.extend(
                                Node(node=route_dependency, parent=p)
                                for route_dependency in route_dependencies
                            )
                            nodes.extend(
                                self._resolve_recursive(route_dependencies, seen=seen)
                            )

                        if parameter not in seen:
                            nodes.extend(self._resolve_recursive(provider, seen=seen))
                # single provider
                else:
                    nodes.append(Node(node=provider, parent=root))
                    # not sure why anyone would ever supply a single APIRouter in an
                    # application, but just in case
                    if isinstance(provider, Supply):
                        route_dependencies = _extract_routes_from_supply(provider)
                        nodes.extend(
                            Node(node=route_dependency, parent=provider)
                            for route_dependency in route_dependencies
                        )
                        nodes.extend(self._resolve_recursive(route_dependencies, seen=seen))
                    if parameter not in seen:
                        nodes.extend(self._resolve_recursive([provider], seen=seen))

                seen.add(parameter)

        return nodes


def _extract_routes_from_supply(supply: Supply) -> list[Dependency]:
    if supply.is_multiprovider:
        inner = supply._value[0]
        if isinstance(inner, APIRouter):
            return [
                APIRouteDependency(route, block_name=supply.block_name)
                for route in inner.routes
                if isinstance(route, APIRoute)
            ]
    return []


class APIRouteDependency(Dependency):
    """
    This is a pseudo-dependency that is only used when calling FastAPIEngin.graph() in
    order to provide richer metadata to the Node.

    This class should never be constructed in application code.
    """

    def __init__(self, route: APIRoute, block_name: str | None = None) -> None:
        """
        Warning: this should never be constructed in application code.
        """
        self._route = route
        self._signature = inspect.signature(route.endpoint)
        self._block_name = block_name

    @property
    def route(self) -> APIRoute:
        return self._route

    @property
    def parameter_types(self) -> list[TypeId]:
        parameters = list(self._signature.parameters.values())
        if not parameters:
            return []
        if parameters[0].name == "self":
            parameters.pop(0)
        return [
            type_id_of(typing.get_args(param.annotation)[0])
            for param in parameters
            if self._is_injected_param(param)
        ]

    @staticmethod
    def _is_injected_param(param: Parameter) -> bool:
        if typing.get_origin(param.annotation) != typing.Annotated:
            return False
        args = typing.get_args(param.annotation)
        if len(args) != 2:
            return False
        return isinstance(args[1], Depends) and hasattr(args[1], "__engin__")

    @property
    def name(self) -> str:
        methods = ",".join(self._route.methods)
        return f"{methods} {self._route.path}"



import asyncio
import logging
from collections import defaultdict
from collections.abc import Collection, Iterable
from dataclasses import dataclass
from inspect import BoundArguments, Signature
from typing import Any, Generic, TypeVar, cast

from engin._dependency import Dependency, Provide, Supply
from engin._exceptions import ProviderError
from engin._type_utils import TypeId, type_id_of

LOG = logging.getLogger("engin")

T = TypeVar("T")


@dataclass(slots=True, kw_only=True, frozen=True)
class AssembledDependency(Generic[T]):
    """
    An AssembledDependency can be called to construct the result.
    """

    dependency: Dependency[Any, T]
    bound_args: BoundArguments

    async def __call__(self) -> T:
        """
        Construct the dependency.

        Returns:
            The constructed value.
        """
        return await self.dependency(*self.bound_args.args, **self.bound_args.kwargs)


class Assembler:
    """
    A container for Providers that is responsible for building provided types.

    The Assembler acts as a cache for previously built types, meaning repeat calls
    to `get` will produce the same value.

    Examples:
        ```python
        def build_str() -> str:
            return "foo"

        a = Assembler([Provide(build_str)])
        await a.get(str)
        ```
    """

    def __init__(self, providers: Iterable[Provide]) -> None:
        self._providers: dict[TypeId, Provide[Any]] = {}
        self._multiproviders: dict[TypeId, list[Provide[list[Any]]]] = defaultdict(list)
        self._dependencies: dict[TypeId, Any] = {}
        self._consumed_providers: set[Provide[Any]] = set()
        self._lock = asyncio.Lock()

        for provider in providers:
            type_id = provider.return_type_id
            if not provider.is_multiprovider:
                if type_id in self._providers:
                    raise RuntimeError(f"A Provider already exists for '{type_id}'")
                self._providers[type_id] = provider
            else:
                self._multiproviders[type_id].append(provider)

    async def assemble(self, dependency: Dependency[Any, T]) -> AssembledDependency[T]:
        """
        Assemble a dependency.

        Given a Dependency type, such as Invoke, the Assembler constructs the types
        required by the Dependency's signature from its providers.

        Args:
            dependency: the Dependency to assemble.

        Returns:
            An AssembledDependency, which can be awaited to construct the final value.
        """
        async with self._lock:
            return AssembledDependency(
                dependency=dependency,
                bound_args=await self._bind_arguments(dependency.signature),
            )

    async def get(self, type_: type[T]) -> T:
        """
        Return the constructed value for the given type.

        This method assembles the required Providers and constructs their corresponding
        values.

        If the

        Args:
            type_: the type of the desired value.

        Raises:
            LookupError: When no provider is found for the given type.
            ProviderError: When a provider errors when trying to construct the type or
                any of its dependent types.

        Returns:
            The constructed value.
        """
        type_id = type_id_of(type_)
        if type_id in self._dependencies:
            return cast(T, self._dependencies[type_id])
        if type_id.multi:
            out = []
            if type_id not in self._multiproviders:
                raise LookupError(f"no provider found for target type id '{type_id}'")
            for provider in self._multiproviders[type_id]:
                assembled_dependency = await self.assemble(provider)
                try:
                    out.extend(await assembled_dependency())
                except Exception as err:
                    raise ProviderError(
                        provider=provider,
                        error_type=type(err),
                        error_message=str(err),
                    ) from err
            self._dependencies[type_id] = out
            return out  # type: ignore[return-value]
        else:
            if type_id not in self._providers:
                raise LookupError(f"no provider found for target type id '{type_id}'")
            assembled_dependency = await self.assemble(self._providers[type_id])
            try:
                value = await assembled_dependency()
            except Exception as err:
                raise ProviderError(
                    provider=self._providers[type_id],
                    error_type=type(err),
                    error_message=str(err),
                ) from err
            self._dependencies[type_id] = value
            return value  # type: ignore[return-value]

    def has(self, type_: type[T]) -> bool:
        return type_id_of(type_) in self._providers

    def _resolve_providers(self, type_id: TypeId) -> Collection[Provide]:
        if type_id.multi:
            providers = self._multiproviders.get(type_id)
        else:
            providers = [provider] if (provider := self._providers.get(type_id)) else None
        if not providers:
            if type_id.multi:
                LOG.warning(f"no provider for '{type_id}' defaulting to empty list")
                providers = [(Supply([], type_hint=list[type_id.type]))]  # type: ignore[name-defined]
            else:
                raise LookupError(f"No Provider registered for dependency '{type_id}'")

        required_providers: list[Provide[Any]] = []
        for provider in providers:
            required_providers.extend(
                provider
                for provider_param in provider.parameter_types
                for provider in self._resolve_providers(provider_param)
            )

        return {*required_providers, *providers}

    async def _satisfy(self, target: TypeId) -> None:
        for provider in self._resolve_providers(target):
            if provider in self._consumed_providers:
                continue
            self._consumed_providers.add(provider)
            type_id = provider.return_type_id
            bound_args = await self._bind_arguments(provider.signature)
            try:
                value = await provider(*bound_args.args, **bound_args.kwargs)
            except Exception as err:
                raise ProviderError(
                    provider=provider, error_type=type(err), error_message=str(err)
                ) from err
            if provider.is_multiprovider:
                if type_id in self._dependencies:
                    self._dependencies[type_id].extend(value)
                else:
                    self._dependencies[type_id] = value
            else:
                self._dependencies[type_id] = value

    async def _bind_arguments(self, signature: Signature) -> BoundArguments:
        args = []
        kwargs = {}
        for param_name, param in signature.parameters.items():
            if param_name == "self":
                args.append(object())
                continue
            param_key = type_id_of(param.annotation)
            has_dependency = param_key in self._dependencies
            if not has_dependency:
                await self._satisfy(param_key)
            val = self._dependencies[param_key]
            if param.kind == param.POSITIONAL_ONLY:
                args.append(val)
            else:
                kwargs[param.name] = val

        return signature.bind(*args, **kwargs)



from typing import Annotated

import pytest

from engin import Assembler, Invoke, Provide
from engin._exceptions import ProviderError
from tests.deps import make_int, make_many_int, make_many_int_alt, make_str


async def test_assembler():
    assembler = Assembler([Provide(make_int), Provide(make_str), Provide(make_many_int)])

    def assert_all(some_int: int, some_str: str, many_ints: list[int]):
        assert isinstance(some_str, str)
        assert isinstance(some_int, int)
        assert all(isinstance(x, int) for x in many_ints)

    assembled_dependency = await assembler.assemble(Invoke(assert_all))

    await assembled_dependency()


async def test_assembler_with_multiproviders():
    assembler = Assembler([Provide(make_many_int), Provide(make_many_int_alt)])

    def assert_all(many_ints: list[int]):
        expected_ints = [*make_many_int(), *make_many_int_alt()]
        assert sorted(many_ints) == sorted(expected_ints)

    assembled_dependency = await assembler.assemble(Invoke(assert_all))

    await assembled_dependency()


async def test_assembler_providers_only_called_once():
    _count = 0

    def count() -> int:
        nonlocal _count
        _count += 1
        return _count

    def assert_singleton(some: int) -> None:
        assert some == 1

    assembler = Assembler([Provide(count)])

    assembled_dependency = await assembler.assemble(Invoke(assert_singleton))
    await assembled_dependency()

    assembled_dependency = await assembler.assemble(Invoke(assert_singleton))
    await assembled_dependency()


async def test_assembler_with_unknown_type_raises_lookup_error():
    assembler = Assembler([])

    with pytest.raises(LookupError):
        await assembler.get(str)


async def test_assembler_with_unknown_type_raises_assembly_error():
    def make_str() -> str:
        raise RuntimeError("foo")

    assembler = Assembler([Provide(make_str)])

    with pytest.raises(ProviderError):
        await assembler.get(str)


async def test_annotations():
    def make_str_1() -> Annotated[str, "1"]:
        return "bar"

    def make_str_2() -> Annotated[str, "2"]:
        return "foo"

    assembler = Assembler([Provide(make_str_1), Provide(make_str_2)])

    with pytest.raises(LookupError):
        await assembler.get(str)

    assert await assembler.get(Annotated[str, "1"]) == "bar"
    assert await assembler.get(Annotated[str, "2"]) == "foo"



import inspect
from collections.abc import Iterable, Iterator
from typing import ClassVar

from engin._dependency import Func, Invoke, Provide


def provide(func: Func) -> Func:
    """
    A decorator for defining a Provider in a Block.
    """
    func._opt = Provide(func)  # type: ignore[attr-defined]
    return func


def invoke(func: Func) -> Func:
    """
    A decorator for defining an Invocation in a Block.
    """
    func._opt = Invoke(func)  # type: ignore[attr-defined]
    return func


class Block(Iterable[Provide | Invoke]):
    """
    A Block is a collection of providers and invocations.

    Blocks are useful for grouping a collection of related providers and invocations, and
    are themselves a valid Option type that can be passed to the Engin.

    Providers are defined as methods decorated with the `provide` decorator, and similarly
    for Invocations and the `invoke` decorator.

    Examples:
        Define a simple block.

        ```python3
        from engin import Block, provide, invoke

        class MyBlock(Block):
            @provide
            def some_str(self) -> str:
                return "foo"

            @invoke
            def print_str(self, string: str) -> None:
                print(f"invoked on string '{string}')
        ```
    """

    options: ClassVar[list[Provide | Invoke]] = []

    def __init__(self, /, block_name: str | None = None) -> None:
        self._options: list[Provide | Invoke] = self.options[:]
        self._name = block_name or f"{type(self).__name__}"
        for _, method in inspect.getmembers(self):
            if opt := getattr(method, "_opt", None):
                if not isinstance(opt, Provide | Invoke):
                    raise RuntimeError("Block option is not an instance of Provide or Invoke")
                opt.set_block_name(self._name)
                self._options.append(opt)
        for opt in self.options:
            opt.set_block_name(self._name)

    @property
    def name(self) -> str:
        return self._name

    def __iter__(self) -> Iterator[Provide | Invoke]:
        return iter(self._options)



from abc import ABC, abstractmethod
from typing import Any


class DatabaseInterface(ABC):
    @abstractmethod
    def get(self, key: str) -> Any: ...

    @abstractmethod
    def list(self) -> Any: ...

    @abstractmethod
    def set(self, key: str, value: Any) -> Any: ...



from engin import Block, Engin, invoke, provide


def test_block():
    class MyBlock(Block):
        @provide
        def provide_int(self) -> int:
            return 3

        @invoke
        def invoke_square(self, some: int) -> None:
            return some**2

    my_block = MyBlock()

    options = list(iter(my_block))
    assert len(options) == 2

    assert Engin(my_block)



from contextlib import asynccontextmanager

from starlette.applications import Starlette

from engin import Engin, Invoke, Lifecycle, Provide
from engin.ext.asgi import ASGIEngin, ASGIType


def a(lifecycle: Lifecycle) -> None:
    @asynccontextmanager
    async def _raise_err() -> None:
        raise RuntimeError("Error in Startup!")
        yield

    lifecycle.append(_raise_err())


B_LIFECYCLE_STATE = False


def b(lifecycle: Lifecycle) -> None:
    @asynccontextmanager
    async def _b_startup() -> None:
        global B_LIFECYCLE_STATE
        B_LIFECYCLE_STATE = True
        yield

    lifecycle.append(_b_startup())


async def test_error_in_startup():
    engin = Engin(Invoke(a), Invoke(b))

    await engin.start()
    assert not B_LIFECYCLE_STATE


async def test_error_in_startup_asgi():
    def asgi_type() -> ASGIType:
        return Starlette()

    engin = ASGIEngin(Invoke(a), Invoke(b), Provide(asgi_type))

    await engin.start()
    assert not B_LIFECYCLE_STATE



from datetime import UTC, datetime
from typing import TypeAlias

from engin import Block, provide


def make_int() -> int:
    return 1


def make_str() -> str:
    return "foo"


def make_many_int() -> list[int]:
    return [2, 3, 4]


def make_many_int_alt() -> list[int]:
    return [5, 6, 7]


IntTypeAlias: TypeAlias = int


def make_aliased_int() -> IntTypeAlias:
    return 8


class ABlock(Block):
    @provide
    def make_datetime(self) -> datetime:
        return datetime.now(tz=UTC)

    @provide
    def make_many_float(self) -> list[float]:
        return [1.2, 2.3]



import asyncio
import logging
import os
import signal
from asyncio import Event, Task
from collections.abc import Iterable
from contextlib import AsyncExitStack
from itertools import chain
from types import FrameType
from typing import ClassVar, TypeAlias

from engin import Entrypoint
from engin._assembler import AssembledDependency, Assembler
from engin._block import Block
from engin._dependency import Dependency, Invoke, Provide, Supply
from engin._graph import DependencyGrapher, Node
from engin._lifecycle import Lifecycle
from engin._type_utils import TypeId

LOG = logging.getLogger("engin")

Option: TypeAlias = Invoke | Provide | Supply | Block
_Opt: TypeAlias = Invoke | Provide | Supply

_OS_IS_WINDOWS = os.name == "nt"


class Engin:
    """
    The Engin is a modular application defined by a collection of options.

    Users should instantiate the Engin with a number of options, where options can be an
    instance of Provide, Invoke, or a collection of these combined in a Block.

    To create a useful application, users should pass in one or more providers (Provide or
    Supply) and at least one invocation (Invoke or Entrypoint).

    When instantiated the Engin can be run. This is typically done via the `run` method,
    but for advanced usecases it can be easier to use the `start` and `stop` methods.

    When ran the Engin takes care of the complete application lifecycle:
    1. The Engin assembles all Invocations. Only Providers that are required to satisfy
       the Invoke options parameters are assembled.
    2. All Invocations are run sequentially in the order they were passed in to the Engin.
    3. Any Lifecycle Startup defined by a provider that was assembled in order to satisfy
       the constructors is ran.
    4. The Engin waits for a stop signal, i.e. SIGINT or SIGTERM.
    5. Any Lifecyce Shutdown task is ran, in the reverse order to the Startup order.

    Examples:
        ```python
        import asyncio

        from httpx import AsyncClient

        from engin import Engin, Invoke, Provide


        def httpx_client() -> AsyncClient:
            return AsyncClient()


        async def main(http_client: AsyncClient) -> None:
            print(await http_client.get("https://httpbin.org/get"))

        engin = Engin(Provide(httpx_client), Invoke(main))

        asyncio.run(engin.run())
        ```
    """

    _LIB_OPTIONS: ClassVar[list[Option]] = [Provide(Lifecycle)]

    def __init__(self, *options: Option) -> None:
        """
        Initialise the class with the provided options.

        Examples:
            >>> engin = Engin(Provide(construct_a), Invoke(do_b), Supply(C()), MyBlock())

        Args:
            *options: an instance of Provide, Supply, Invoke, Entrypoint or a Block.
        """
        self._stop_requested_event = Event()
        self._stop_complete_event = Event()
        self._exit_stack: AsyncExitStack = AsyncExitStack()
        self._shutdown_task: Task | None = None
        self._run_task: Task | None = None

        # TODO: refactor _destruct_options and related attributes into a dedicated class?
        self._providers: dict[TypeId, Provide] = {TypeId.from_type(Engin): Provide(self._self)}
        self._multiproviders: dict[TypeId, list[Provide]] = {}
        self._invocations: list[Invoke] = []
        # populates the above
        self._destruct_options(chain(self._LIB_OPTIONS, options))
        multi_providers = [p for multi in self._multiproviders.values() for p in multi]
        self._assembler = Assembler(chain(self._providers.values(), multi_providers))

    @property
    def assembler(self) -> Assembler:
        return self._assembler

    async def run(self) -> None:
        """
        Run the engin.

        The engin will run until it is stopped via an external signal (i.e. SIGTERM or
        SIGINT) or the `stop` method is called on the engin.
        """
        await self.start()
        self._run_task = asyncio.create_task(_wait_for_stop_signal(self._stop_requested_event))
        await self._stop_requested_event.wait()
        await self._shutdown()

    async def start(self) -> None:
        """
        Start the engin.

        This is an alternative to calling `run`. This method waits for the startup
        lifecycle to complete and then returns. The caller is then responsible for
        calling `stop`.
        """
        LOG.info("starting engin")
        assembled_invocations: list[AssembledDependency] = [
            await self._assembler.assemble(invocation) for invocation in self._invocations
        ]

        for invocation in assembled_invocations:
            try:
                await invocation()
            except Exception as err:
                name = invocation.dependency.name
                LOG.error(f"invocation '{name}' errored, exiting", exc_info=err)
                return

        lifecycle = await self._assembler.get(Lifecycle)

        try:
            for hook in lifecycle.list():
                await self._exit_stack.enter_async_context(hook)
        except Exception as err:
            LOG.error("lifecycle startup error, exiting", exc_info=err)
            await self._exit_stack.aclose()
            return

        LOG.info("startup complete")

        self._shutdown_task = asyncio.create_task(self._shutdown_when_stopped())

    async def stop(self) -> None:
        """
        Stop the engin.

        This method will wait for the shutdown lifecycle to complete before returning.
        Note this method can be safely called at any point, even before the engin is
        started.
        """
        self._stop_requested_event.set()
        if self._shutdown_task is None:
            return
        await self._stop_complete_event.wait()

    def graph(self) -> list[Node]:
        grapher = DependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve(self._invocations)

    async def _shutdown(self) -> None:
        LOG.info("stopping engin")
        await self._exit_stack.aclose()
        self._stop_complete_event.set()
        LOG.info("shutdown complete")

    async def _shutdown_when_stopped(self) -> None:
        await self._stop_requested_event.wait()
        await self._shutdown()

    def _destruct_options(self, options: Iterable[Option]) -> None:
        for opt in options:
            if isinstance(opt, Block):
                self._destruct_options(opt)
            if isinstance(opt, Provide | Supply):
                if not opt.is_multiprovider:
                    existing = self._providers.get(opt.return_type_id)
                    self._log_option(opt, overwrites=existing)
                    self._providers[opt.return_type_id] = opt
                else:
                    self._log_option(opt)
                    if opt.return_type_id in self._multiproviders:
                        self._multiproviders[opt.return_type_id].append(opt)
                    else:
                        self._multiproviders[opt.return_type_id] = [opt]
            elif isinstance(opt, Invoke):
                self._log_option(opt)
                self._invocations.append(opt)

    @staticmethod
    def _log_option(opt: Dependency, overwrites: Dependency | None = None) -> None:
        if overwrites is not None:
            extra = f"\tOVERWRITES {overwrites.name}"
            if overwrites.block_name:
                extra += f" [{overwrites.block_name}]"
        else:
            extra = ""
        if isinstance(opt, Supply):
            LOG.debug(f"SUPPLY      {opt.return_type_id!s:<35}{extra}")
        elif isinstance(opt, Provide):
            LOG.debug(f"PROVIDE     {opt.return_type_id!s:<35} <- {opt.name}() {extra}")
        elif isinstance(opt, Entrypoint):
            type_id = opt.parameter_types[0]
            LOG.debug(f"ENTRYPOINT  {type_id!s:<35}")
        elif isinstance(opt, Invoke):
            LOG.debug(f"INVOKE      {opt.name:<35}")

    def _self(self) -> "Engin":
        return self


async def _wait_for_stop_signal(stop_requested_event: Event) -> None:
    try:
        # try to gracefully handle sigint/sigterm
        if not _OS_IS_WINDOWS:
            loop = asyncio.get_running_loop()
            for signame in (signal.SIGINT, signal.SIGTERM):
                loop.add_signal_handler(signame, stop_requested_event.set)

            await stop_requested_event.wait()
        else:
            should_stop = False

            # windows does not support signal_handlers, so this is the workaround
            def ctrlc_handler(sig: int, frame: FrameType | None) -> None:
                nonlocal should_stop
                if should_stop:
                    raise KeyboardInterrupt("Forced keyboard interrupt")
                should_stop = True

            signal.signal(signal.SIGINT, ctrlc_handler)

            while not should_stop:
                # In case engin is stopped via external `stop` call.
                if stop_requested_event.is_set():
                    return
                await asyncio.sleep(0.1)

            stop_requested_event.set()
    except asyncio.CancelledError:
        pass



import asyncio
from collections.abc import Iterable
from contextlib import asynccontextmanager
from datetime import datetime

import pytest

from engin import Engin, Entrypoint, Invoke, Lifecycle, Provide, ProviderError
from tests.deps import ABlock


class A:
    def __init__(self): ...


class B:
    def __init__(self): ...


class C:
    def __init__(self): ...


async def test_engin():
    def a() -> A:
        return A()

    def b(_: A) -> B:
        return B()

    def c(_: B) -> C:
        return C()

    def multi_a() -> list[A]:
        return [A()]

    def multi_a_2() -> list[A]:
        return [A(), A()]

    def main(c: C, multi_a: list[A]) -> None:
        assert isinstance(c, C)
        assert len(multi_a) == 3

    engin = Engin(
        Provide(a), Provide(b), Provide(c), Provide(multi_a), Provide(multi_a_2), Invoke(main)
    )

    await engin.start()
    await engin.stop()


async def test_engin_with_block():
    def main(dt: datetime, floats: list[float]) -> None:
        assert isinstance(dt, datetime)
        assert isinstance(floats, list)
        assert all(isinstance(x, float) for x in floats)

    engin = Engin(ABlock(), Invoke(main))

    await engin.start()
    await engin.stop()


async def test_engin_error_handling():
    async def raise_value_error() -> int:
        raise ValueError("foo")

    async def main(foo: int) -> None:
        return

    engin = Engin(Provide(raise_value_error), Invoke(main))

    with pytest.raises(ProviderError, match="foo"):
        await engin.run()


async def test_engin_with_entrypoint():
    provider_called = False

    def a() -> A:
        nonlocal provider_called
        provider_called = True
        return A()

    engin = Engin(Provide(a), Entrypoint(A))

    await engin.start()
    await engin.stop()

    assert provider_called


async def test_engin_with_lifecycle():
    state = 0

    @asynccontextmanager
    async def lifespan_task() -> Iterable[None]:
        nonlocal state
        state = 1
        yield
        state = 2

    def foo(lifecycle: Lifecycle) -> None:
        lifecycle.append(lifespan_task())

    engin = Engin(Invoke(foo))

    await engin.start()
    assert state == 1

    await engin.stop()
    assert state == 2


async def test_engin_with_lifecycle_using_run():
    state = 0

    @asynccontextmanager
    async def lifespan_task() -> Iterable[None]:
        nonlocal state
        state = 1
        yield
        state = 2

    def foo(lifecycle: Lifecycle) -> None:
        lifecycle.append(lifespan_task())

    engin = Engin(Invoke(foo))

    async def _stop_task():
        await asyncio.sleep(0.25)
        # lifecycle should have started by now
        assert state == 1
        await engin.stop()

    await asyncio.gather(engin.run(), _stop_task())
    # lifecycle should have stopped by now
    assert state == 2


def test_engin_graph():
    def a() -> A:
        return A()

    def b(_: A) -> B:
        return B()

    def c(_: B) -> C:
        return C()

    def main(c: C) -> None:
        assert isinstance(c, C)

    engin = Engin(Provide(a), Provide(b), Provide(c), Invoke(main))

    graph = engin.graph()

    assert len(graph) == 3



from examples.fastapi.routes.cats.domain import Cat
from examples.fastapi.routes.cats.ports import CatRepository


class InMemoryCatRepository(CatRepository):
    def __init__(self) -> None:
        self._cats: dict[int, Cat] = {}

    def get(self, cat_id: int) -> Cat:
        if cat_id in self._cats:
            return self._cats[cat_id]
        raise LookupError(f"No cat found for id: {cat_id}")

    def set(self, cat: Cat) -> None:
        self._cats[cat.id] = cat

    def next_id(self) -> int:
        return len(self._cats)



import importlib
import logging
import socketserver
import sys
import threading
from argparse import ArgumentParser
from http.server import BaseHTTPRequestHandler
from time import sleep
from typing import Any

from engin import Engin, Entrypoint, Invoke
from engin._dependency import Dependency, Provide, Supply
from engin.ext.asgi import ASGIEngin
from engin.ext.fastapi import APIRouteDependency

# mute logging from importing of files + engin's debug logging.
logging.disable()

args = ArgumentParser(
    prog="engin-graph",
    description="Creates a visualisation of your application's dependencies",
)
args.add_argument(
    "app",
    help=(
        "the import path of your Engin instance, in the form "
        "'package:application', e.g. 'app.main:engin'"
    ),
)


def serve_graph() -> None:
    # add cwd to path to enable local package imports
    sys.path.insert(0, "")

    parsed = args.parse_args()

    app = parsed.app

    try:
        module_name, engin_name = app.split(":", maxsplit=1)
    except ValueError:
        raise ValueError(
            "Expected an argument of the form 'module:attribute', e.g. 'myapp:engin'"
        ) from None

    module = importlib.import_module(module_name)

    try:
        instance = getattr(module, engin_name)
    except LookupError:
        raise LookupError(f"Module '{module_name}' has no attribute '{engin_name}'") from None

    if not isinstance(instance, Engin):
        raise TypeError(f"'{app}' is not an Engin instance")

    nodes = instance.graph()

    # transform dependencies into mermaid syntax
    dependencies = [
        f"{_render_node(node.parent)} --> {_render_node(node.node)}"
        for node in nodes
        if node.parent is not None
    ]

    html = (
        _GRAPH_HTML.replace("%%DATA%%", "\n".join(dependencies))
        .replace(
            "%%LEGEND%%",
            ASGI_ENGIN_LEGEND if isinstance(instance, ASGIEngin) else DEFAULT_LEGEND,
        )
        .encode("utf8")
    )

    class Handler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            self.send_response(200, "OK")
            self.send_header("Content-type", "html")
            self.end_headers()
            self.wfile.write(html)

        def log_message(self, format: str, *args: Any) -> None:
            return

    def _start_server() -> None:
        with socketserver.TCPServer(("localhost", 8123), Handler) as httpd:
            print("Serving dependency graph on http://localhost:8123")
            httpd.serve_forever()

    server_thread = threading.Thread(target=_start_server)
    server_thread.daemon = True  # Daemonize the thread so it exits when the main script exits
    server_thread.start()

    try:
        sleep(10000)
    except KeyboardInterrupt:
        print("Exiting the server...")


_BLOCK_IDX: dict[str, int] = {}
_SEEN_BLOCKS: list[str] = []


def _render_node(node: Dependency) -> str:
    node_id = id(node)
    md = ""
    style = ""

    # format block name
    if n := node.block_name:
        md += f"_{n}_\n"
        if n not in _BLOCK_IDX:
            _BLOCK_IDX[n] = len(_SEEN_BLOCKS) % 8
            _SEEN_BLOCKS.append(n)
        style = f":::b{_BLOCK_IDX[n]}"

    if isinstance(node, Supply):
        md += f"{node.return_type_id}"
        return f'{node_id}("`{md}`"){style}'
    if isinstance(node, Provide):
        md += f"{node.return_type_id}"
        return f'{node_id}["`{md}`"]{style}'
    if isinstance(node, Entrypoint):
        entrypoint_type = node.parameter_types[0]
        md += f"{entrypoint_type}"
        return f'{node_id}[/"`{md}`"\\]{style}'
    if isinstance(node, Invoke):
        md += f"{node.func_name}"
        return f'{node_id}[/"`{md}`"/]{style}'
    if isinstance(node, APIRouteDependency):
        md += f"{node.name}"
        return f'{node_id}[["`{md}`"]]{style}'
    else:
        return f'{node_id}["`{node.name}`"]{style}'


_GRAPH_HTML = """
<!doctype html>
<html lang="en">
  <body>
    <div style="border-style:outset">
        <p>LEGEND</p>
        <pre class="mermaid">
          graph LR
            %%LEGEND%%
            classDef b0 fill:#7fc97f;
        </pre>
    </div>
    <pre class="mermaid">
      graph TD
          %%DATA%%
          classDef b0 fill:#7fc97f;
          classDef b1 fill:#beaed4;
          classDef b2 fill:#fdc086;
          classDef b3 fill:#ffff99;
          classDef b4 fill:#386cb0;
          classDef b5 fill:#f0027f;
          classDef b6 fill:#bf5b17;
          classDef b7 fill:#666666;
    </pre>
    <script type="module">
      import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
      let config = { flowchart: { useMaxWidth: false, htmlLabels: true } };
      mermaid.initialize(config);
    </script>
  </body>
</html>
"""

DEFAULT_LEGEND = (
    "0[/Invoke/] ~~~ 1[/Entrypoint\\] ~~~ 2[Provide] ~~~ 3(Supply)"
    ' ~~~ 4["`Block Grouping`"]:::b0'
)
ASGI_ENGIN_LEGEND = DEFAULT_LEGEND + " ~~~ 5[[API Route]]"



import traceback
from collections.abc import Awaitable, Callable, MutableMapping
from typing import Any, ClassVar, Protocol, TypeAlias

from engin import Engin, Entrypoint, Option

__all__ = ["ASGIEngin", "ASGIType"]

from engin._graph import DependencyGrapher, Node

_Scope: TypeAlias = MutableMapping[str, Any]
_Message: TypeAlias = MutableMapping[str, Any]
_Receive: TypeAlias = Callable[[], Awaitable[_Message]]
_Send: TypeAlias = Callable[[_Message], Awaitable[None]]


class ASGIType(Protocol):
    async def __call__(self, scope: _Scope, receive: _Receive, send: _Send) -> None: ...


class ASGIEngin(Engin, ASGIType):
    _asgi_type: ClassVar[type[ASGIType]] = ASGIType  # type: ignore[type-abstract]
    _asgi_app: ASGIType

    def __init__(self, *options: Option) -> None:
        super().__init__(*options)

        if not self._assembler.has(self._asgi_type):
            raise LookupError(
                f"A provider for `{self._asgi_type.__name__}` was expected, none found"
            )

    async def __call__(self, scope: _Scope, receive: _Receive, send: _Send) -> None:
        if scope["type"] == "lifespan":
            message = await receive()
            receive = _Rereceive(message)
            if message["type"] == "lifespan.startup":
                try:
                    await self._startup()
                except Exception as err:
                    exc = "".join(traceback.format_exception(err))
                    await send({"type": "lifespan.startup.failed", "message": exc})

            elif message["type"] == "lifespan.shutdown":
                await self.stop()

        await self._asgi_app(scope, receive, send)

    async def _startup(self) -> None:
        await self.start()
        self._asgi_app = await self._assembler.get(self._asgi_type)

    def graph(self) -> list[Node]:
        grapher = DependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve([Entrypoint(self._asgi_type), *self._invocations])


class _Rereceive:
    def __init__(self, message: _Message) -> None:
        self._message = message

    async def __call__(self, *args: Any, **kwargs: Any) -> _Message:
        return self._message
